# This code implements a clone of the asyncio mainloop which hooks into
# Trio.

import sys
import trio
import asyncio
import math
import heapq
import signal
import threading
import selectors

import logging
logger = logging.getLogger(__name__)

from functools import partial
from asyncio.events import _format_callback, _get_function_source
from selectors import _BaseSelectorImpl, EVENT_READ, EVENT_WRITE

__all__ = ['TrioEventLoop']


class _Clear:
    def clear(self):
        pass


def _format_callback_source(func, args, kwargs):
    func_repr = _format_callback(func, args, kwargs)
    source = _get_function_source(func)
    if source:  # pragma: no cover
        func_repr += ' at %s:%s' % source
    return func_repr


class _QuitEvent(trio.Event):
    """Used to signal the mainloop to stop,
    as opposed to a "normal" event used for syncing.
    """
    pass


class _TrioHandle:
    """
    This extends asyncio.Handle by providing:
    * a way to pass keyword arguments
    * a way to cancel an async callback
    * a way to declare the type of the callback function
    
    ``is_sync`` may be
    * True: sync function, use _call_sync()
    * False: async function, use _call_async()
    * None: also async, but the callback function accepts
      the handle as its sole argument

    The caller is responsible for checking whether the handle
    has been cancelled before invoking ``call_[a]sync()``.
    """

    def _init(self, kwargs, is_sync):
        """Secondary init.
        """
        self._kwargs = kwargs
        self._is_sync = is_sync
        self._scope = None

    def cancel(self):
        super().cancel()
        self._kwargs = None
        if self._scope is not None:
            self._scope.cancel()

    def _cb_future_cancel(self, f):
        """If a Trio task completes an asyncio Future,
        add this callback to the future
        and set ``_scope`` to the Trio cancel scope
        so that the task is terminated when the future gets canceled.

        """
        if f.cancelled():
            self.cancel()

    def _repr_info(self):
        info = [self.__class__.__name__]
        if self._cancelled:
            info.append('cancelled')
        if self._callback is not None:
            info.append(
                _format_callback_source(
                    self._callback, self._args, self._kwargs
                )
            )
        if self._source_traceback:
            frame = self._source_traceback[-1]
            info.append('created at %s:%s' % (frame[0], frame[1]))
        return info

    def _call_sync(self):
        assert self._is_sync
        assert not self._cancelled
        res = self._callback(*self._args, **self._kwargs)
        return res

    async def _call_async(self):
        assert not self._is_sync
        assert not self._cancelled
        try:
            with trio.open_cancel_scope() as scope:
                self._scope = scope
                if self._is_sync is None:
                    res = await self._callback(self)
                else:
                    res = await self._callback(*self._args, **self._kwargs)
            return res
        finally:
            self._scope = None


class Handle(_TrioHandle, asyncio.Handle):
    def __init__(self, callback, args, kwargs, loop, is_sync):
        super().__init__(callback, args, loop)
        self._init(kwargs, is_sync)


class DeltaTime:
    __slots__ = ('delta')

    def __init__(self, delta=0):
        self.delta = delta

    def __add__(self, x):
        return DeltaTime(self.delta + x)

    def __iadd__(self, x):
        self.delta += x
        return self

    def __sub__(self, x):
        if isinstance(x, DeltaTime):
            return self.delta - x.delta
        return DeltaTime(self.delta - x)

    def __isub__(self, x):
        self.delta -= x
        return self


class TimerHandle(_TrioHandle, asyncio.TimerHandle):
    def __init__(
            self, when, callback, args, kwargs, loop, is_sync,
            is_relative=False
    ):
        super().__init__(when, callback, args, loop)
        if isinstance(when, DeltaTime):
            assert not is_relative
            self._when = when.delta
            is_relative = True
        self._init(kwargs, is_sync)
        self._relative = is_relative

    def _abs_time(self):
        if self._relative:  # pragma: no branch
            self._when += self._loop.time()
            self._relative = False

    def _rel_time(self):
        if not self._relative:  # pragma: no branch
            self._when -= self._loop.time()
            self._relative = True


class _TrioSelector(_BaseSelectorImpl):
    """A selector that hooks into a ``TrioEventLoop``.

    In fact it's just a basic selector that disables the actual
    ``select()`` method, as that is controlled by the event loop.
    """

    def select(self, timeout=None):  # pragma: no cover
        raise NotImplementedError

    def _select(self, r, w, x, timeout=None):  # pragma: no cover
        raise NotImplementedError


class TrioExecutor:
    def __init__(self, limiter=None):
        self._running = True
        self._limiter = limiter

    async def submit(self, func, *args):
        if not self._running:  # pragma: no cover
            raise RuntimeError("Executor is down")
        return await trio.run_sync_in_worker_thread(
            func, *args, limiter=self._limiter
        )

    def shutdown(self, wait=None):
        self._running = False


class TrioEventLoop(asyncio.unix_events._UnixSelectorEventLoop):
    """An asyncio mainloop for trio

    This code implements a semi-efficient way to run asyncio code within Trio.
    """
    # for calls from other threads
    _token = None

    # to propagate exceptions raised in the main loop
    _exc = None

    def __init__(self):
        # Processing queue
        self._q = trio.Queue(9999)

        # interim storage for calls to ``call_soon_threadsafe``
        # while the main loop is not running
        self._delayed_calls = []

        super().__init__(_TrioSelector())

        # replaced internal data
        self._ready = _Clear()
        self._scheduled = _Clear()
        self._default_executor = TrioExecutor()

        self._orig_signals = {}

        # we need to do our own timeout handling
        self._timers = []

    def time(self):
        """Trio's idea of the current time.

        Trio returns an error when you try to access the current time when
        the asyncio loop is not running. In that case, this function
        returns a DeltaTime object, which represents the time at which
        the loop is restarted.
        """
        if self._token is None:
            return DeltaTime()
        else:
            return trio.current_time()

    async def wait_for(self, future, _scope=None):
        """Wait for an asyncio future in Trio code.

        Cancellations will be propagated bidirectionally;
        the default for propagating asyncio cancellations
        to Trio is the inner-most cancel scope. If that's
        not what you need, pass in an explicit scope`.
        """
        current_task = trio.hazmat.current_task()
        assert self._task is not current_task

        def is_done(f):
            if f.cancelled():
                _scope.cancel()
                return

            exc = f.exception()
            if exc is None:
                res = trio.hazmat.Value(f.result())
            elif isinstance(exc, asyncio.CancelledError):
                _scope.cancel()
                return
            else:
                res = trio.hazmat.Error(exc)
            trio.hazmat.reschedule(current_task, next_send=res)

        def is_aborted(raise_cancel):
            if not future.cancelled():
                future.remove_done_callback(is_done)
                future.cancel()
            return trio.hazmat.Abort.SUCCEEDED

        future.add_done_callback(is_done)
        if _scope is None:  # pragma: no cover
            # TODO: just forbid calling without a scope?
            _scope = trio.hazmat.current_task()._cancel_stack[-1]
        return await trio.hazmat.wait_task_rescheduled(is_aborted)

    async def call_asyncio(self, p, *a, _scope=None, **k):
        """Call an asyncio function or method from Trio.
        
        Returns/Raises: whatever the procedure does.

        Cancellations will be propagated bidirectionally.
        """
        if _scope is None:
            _scope = trio.hazmat.current_task()._cancel_stack[-1]

        try:
            f = asyncio.ensure_future(p(*a), loop=self)
        except asyncio.CancelledError:
            _scope.cancel()
            await trio.sleep(0)
            raise RuntimeError("cancel didn't kill me")  # pragma: no cover

        return await self.wait_for(f, _scope)

    def call_trio(self, p, *a, **k):
        """Call an asynchronous Trio function from asyncio.

        Returns a Future with the result / exception.

        Cancelling the future will cancel the task running your procedure,
        or prevent it from starting if that is stil possible.
        """
        f = asyncio.Future(loop=self)
        h = Handle(self.__call_trio, (
            f,
            p,
        ) + a, k, self, None)
        self._queue_handle(h)
        f.add_done_callback(h._cb_future_cancel)
        return f

    async def __call_trio(self, h):
        f, proc, *args = h._args
        if f.cancelled():  # pragma: no cover
            return
        try:
            with trio.open_cancel_scope() as scope:
                h._scope = scope
                res = await proc(*args, **h._kwargs)
            if scope.cancelled_caught:
                f.cancel()
                return
        except Exception as exc:
            if not f.cancelled():  # pragma: no branch
                f.set_exception(exc)
        else:
            if not f.cancelled():  # pragma: no branch
                f.set_result(res)

    def call_trio_sync(self, p, *a, **k):
        """Call a synchronous function from asyncio.

        Returns a Future with the result / exception.

        Cancelling the future will prevent the code from running,
        assuming that is still possible.

        You might need to use this method if your code needs access to
        features which are only available when Trio is running, such as
        global task-specific variables or the current time.
        (Otherwise you should simply call the code in question directly.)
        """
        f = asyncio.Future(loop=self)
        self._queue_handle(
            Handle(self.__call_trio_sync, (
                f,
                p,
            ) + a, k, self, None)
        )
        return f

    async def __call_trio_sync(self, h):
        f, proc, *args = h._args
        if f.cancelled():  # pragma: no cover
            return
        try:
            res = proc(*args, **h._kwargs)
        except trio.Cancelled:  # pragma: no cover
            raise RuntimeError("Can't cancel sync code")
        except asyncio.CancelledError:  # pragma: no cover
            f.cancel()
            raise
        except Exception as exc:
            f.set_exception(exc)
        else:
            f.set_result(res)

    def call_later(self, delay, callback, *args, **kwargs):
        """asyncio's timer-based delay

        Note that the callback is a sync function.
        """
        self._check_callback(callback, 'call_later')
        self._check_closed()
        assert delay >= 0, delay
        h = TimerHandle(delay, callback, args, kwargs, self, True, True)
        if self._token is None:
            self._delayed_calls.append(h)
        else:
            h = TimerHandle(delay, callback, args, kwargs, self, True, True)
            self._q.put_nowait(h)
        return h

    def _queue_handle(self, handle):
        if self._token is None:
            self._delayed_calls.append(handle)
        else:
            self._q.put_nowait(handle)
        return handle

    def call_at(self, when, callback, *args, **kwargs):
        """asyncio's time-based delay

        Note that the callback is a sync function.
        """
        self._check_callback(callback, 'call_at')
        self._check_closed()
        return self._queue_handle(
            TimerHandle(when, callback, args, kwargs, self, True)
        )

    def call_soon(self, callback, *args, **kwargs):
        self._check_callback(callback, 'call_soon')
        self._check_closed()
        return self._queue_handle(Handle(callback, args, kwargs, self, True))

    def call_soon_threadsafe(self, callback, *args, **kwargs):
        self._check_callback(callback, 'call_soon_threadsafe')
        self._check_closed()
        h = Handle(callback, args, kwargs, self, True)
        if self._token is None:
            self._delayed_calls.append(h)
        else:
            self._token.run_sync_soon(self._q.put_nowait, h)
        return h

    # supersede some built-ins which should not be used

    def _add_callback(self, handle, _via_token=False):
        raise RuntimeError("_add_callback() should always be superseded")

    def _add_callback_signalsafe(self, handle):  # pragma: no cover
        raise RuntimeError("_add_callback_signalsafe() should always be superseded")

    def _handle_signal(self, signum):
        raise RuntimeError("_handle_signal() should always be superseded")

    def _timer_handle_cancelled(self, handle):
        pass

    def run_in_executor(self, executor, func, *args, **kwargs):
        """
        Delegate running a synchronous function to another thread.

        Limitation:
        The executor must be None, or a (subclass of) ``TrioExecutor``.

        Returns an asyncio.Future.
        """
        self._check_callback(func, 'run_in_executor')
        self._check_closed()
        if executor is None: # pragma: no branch
            executor = self._default_executor
        assert isinstance(executor, TrioExecutor)
        if kwargs:
            func = patial(func, **kwargs)
        return self.call_trio(executor.submit, func, *args)

    def _handle_sig(self, sig, _):
        h = self._signal_handlers[sig]
        if self._token is None:
            self._delayed_calls.append(h)
        else:
            self._token.run_sync_soon(self._q.put_nowait, h)

    async def _sync(self):
        w = trio.Event()
        self._q.put_nowait(w)
        await w.wait()

    def add_signal_handler(self, sig, callback, *args, **kwargs):
        self._check_signal(sig)
        self._check_closed()
        if sig == signal.SIGKILL:
            raise RuntimeError("SIGKILL cannot be caught")
        h = Handle(callback, args, kwargs, self, True)
        assert sig not in self._signal_handlers, "Signal %d is already caught" % (
            sig,
        )
        self._orig_signals[sig] = signal.signal(sig, self._handle_sig)
        self._signal_handlers[sig] = h

    def remove_signal_handler(self, sig):
        self._check_signal(sig)
        try:
            h = self._signal_handlers.pop(sig)
        except KeyError:
            return False
        h.cancel()
        signal.signal(sig, self._orig_signals[sig])
        del self._orig_signals[sig]
        return True

    def add_reader(self, fd, callback, *args, **kwargs):
        self._ensure_fd_no_transport(fd)
        return self._add_reader(fd, callback, *args, **kwargs)

    def _add_reader(self, fd, callback, *args, **kwargs):
        self._check_closed()
        handle = Handle(callback, args, kwargs, self, True)
        reader = self._set_read_handle(fd, handle)
        if reader is not None:
            reader.cancel()
        if self._token is None:
            return
        self._nursery.start_soon(self._reader_loop, fd, handle)

    def _set_read_handle(self, fd, handle):
        try:
            key = self._selector.get_key(fd)
        except KeyError:
            self._selector.register(fd, EVENT_READ, (handle, None))
            return None
        else:
            mask, (reader, writer) = key.events, key.data
            self._selector.modify(fd, mask | EVENT_READ, (handle, writer))
            return reader

    async def _reader_loop(self, fd, handle, task_status=trio.STATUS_IGNORED):
        task_status.started()
        with trio.open_cancel_scope() as scope:
            handle._scope = scope
            try:
                while not handle._cancelled:  # pragma: no branch
                    await trio.hazmat.wait_readable(fd)
                    handle._call_sync()
                    await self._sync()
            finally:
                handle._scope = None

    # writing to a file descriptor

    def add_writer(self, fd, callback, *args, **kwargs):
        self._ensure_fd_no_transport(fd)
        return self._add_writer(fd, callback, *args, **kwargs)

    def _add_writer(self, fd, callback, *args, **kwargs):
        self._check_closed()
        handle = Handle(callback, args, kwargs, self, True)
        writer = self._set_write_handle(fd, handle)
        if writer is not None:
            writer.cancel()
        if self._token is None:
            return
        self._nursery.start_soon(self._writer_loop, fd, handle)

    def _set_write_handle(self, fd, handle):
        try:
            key = self._selector.get_key(fd)
        except KeyError:
            self._selector.register(fd, EVENT_WRITE, (None, handle))
        else:
            mask, (reader, writer) = key.events, key.data
            self._selector.modify(fd, mask | EVENT_WRITE, (reader, handle))
            return writer

    async def _writer_loop(self, fd, handle, task_status=trio.STATUS_IGNORED):
        with trio.open_cancel_scope() as scope:
            handle._scope = scope
            task_status.started()
            try:
                while not handle._cancelled:  # pragma: no branch
                    await trio.hazmat.wait_writable(fd)
                    handle._call_sync()
                    await self._sync()
            finally:
                handle._scope = None

    def _save_fds(self):
        map = self._selector.get_map()
        saved = [{}, {}]
        for fd, key in list(self._selector.get_map().items()):
            for flag in (0, 1):
                if key.events & (1 << flag):
                    handle = key.data[flag]
                    assert handle is not None
                    if not handle._cancelled:  # pragma: no branch
                        if handle._scope is not None:
                            handle._scope.cancel()

    async def _restore_fds(self):
        for fd, key in list(self._selector.get_map().items()):
            for flag in (0, 1):
                if key.events & (1 << flag):
                    handle = key.data[flag]
                    assert handle is not None
                    if not handle._cancelled:  # pragma: no branch
                        if flag:
                            await self._nursery.start(
                                self._writer_loop, fd, handle
                            )
                        else:
                            await self._nursery.start(
                                self._reader_loop, fd, handle
                            )

    # Trio-based main loop

    async def main_loop(self, task_status=trio.STATUS_IGNORED):
        """This is the Trio replacement of the asyncio loop's main loop.

        Run this method as a standard Trio thread if your main code is
        Trio-based and you need to call asyncio code.

        Use ``run_forever()`` or ``run_until_complete()`` instead if your
        main code is asyncio-based.
        """
        try:
            self._task = trio.hazmat.current_task()

            async with trio.open_nursery() as nursery:
                self._nursery = nursery
                self._token = trio.hazmat.current_trio_token()
                await self._restore_fds()

                try:
                    for obj in self._delayed_calls:
                        if not getattr(obj, '_cancelled', False):
                            if isinstance(obj, TimerHandle):
                                obj._abs_time()
                                heapq.heappush(self._timers, obj)
                            else:
                                self._q.put_nowait(obj)
                    self._delayed_calls = []
                    task_status.started()

                    time_valid = False
                    while True:
                        obj = None
                        if not time_valid:
                            t = self.time()
                            time_valid = True
                        if not self._timers:
                            timeout = math.inf
                        else:
                            timeout = self._timers[0]._when - t
                            if timeout <= 0:
                                obj = heapq.heappop(self._timers)
                        if obj is None:
                            time_valid = False
                            with trio.move_on_after(timeout) as cancel_scope:
                                obj = await self._q.get()
                            if cancel_scope.cancel_called:
                                continue

                            if isinstance(obj, trio.Event):
                                if isinstance(obj, _QuitEvent):
                                    break
                                obj.set()
                                continue
                            if isinstance(obj, TimerHandle):
                                obj._abs_time()
                                heapq.heappush(self._timers, obj)
                                continue
                        if obj._cancelled:
                            continue
                        if getattr(obj, '_is_sync', True) is True:
                            obj._callback(
                                *obj._args, **getattr(obj, '_kwargs', {})
                            )
                        else:
                            nursery.start_soon(obj._call_async)
                except Exception as exc:
                    # The asyncio mainloop would swallow this
                    self._exc = exc
                    return

                finally:
                    # Save open file descriptors
                    try:
                        self._save_fds()
                    except AttributeError:  ## pragma: no cover
                        pass

                    # save timers, by converting them back to relative time
                    for tm in self._timers:
                        tm._rel_time()
                        self._delayed_calls.append(tm)
                    self._timers.clear()

                    del self._nursery
                    self._token = None
                    if isinstance(obj, trio.Event):
                        obj.set()
                    self._stopping = True

        except BaseException as exc:
            print(*trio.format_exception(type(exc), exc, exc.__traceback__))

    def run_task(self, proc, *a, **k):
        """Run a Trio task.

        The asyncio main loop is running in parallel.
        It is stopped when your task finishes.
        """

        if self.is_running():
            raise RuntimeError("This loop is already running")
        f = asyncio.Future(loop=self)
        h = Handle(self.__run_task, (proc, f, a, k), {}, self, False)
        self._delayed_calls.append(h)
        self.run_forever()
        return f.result()

    async def __run_task(self, proc, f, a, k):
        try:
            res = await proc(*a, **k)
        except Exception as exc:
            f.set_exception(exc)
        except trio.Cancelled as exc:  # pragma: no cover
            # No, I am not going to write a test for this
            # Trio is likely to catch the problem internally
            # but better safe than sorry
            try:
                raise RuntimeError("Did you cancel the wrong scope?") from exc
            except RuntimeError as exc:
                if not f.cancelled():
                    f.set_exception(res)
                raise
        else:
            f.set_result(res)
        finally:
            self.stop()

    def run_forever(self):
        """Start the main loop

        This method simply runs ``.main_loop()`` as a trio-enabled version
        of the asyncio main loop.

        Use this (usually via ``run_until_complete()``) if your code
        is asyncio-based and you want to use some trio features or
        libraries.

        Use ``main_loop()`` instead if you main code is trio-based.

        Asyncio's ``run_forever()`` does quite a few setup/teardown things.
        Thus, rather than re-implement all of them, our _run_once() method
        actually implements the main event loop instead of just
        single-stepping.
        """
        super().run_forever()

    def _run_once(self):
        trio.run(self.main_loop)

        # An exception in the main loop needs to be propagated
        # (but only once)
        exc,self._exc = self._exc,None
        if exc is not None:
            raise exc

    def stop(self):
        """Halt the main loop.

        This returns a trio.Event which will trigger when the loop has
        terminated.
        """
        self._stopping = True
        # reset to False by asyncio's run_forever()

        e = _QuitEvent()
        self._queue_handle(e)
        return e

    def close(self):
        forgot_stop = self.is_running()
        if forgot_stop:
            raise RuntimeError("You need to stop the loop before closing it")

        super().close()


class TrioPolicy(asyncio.unix_events._UnixDefaultEventLoopPolicy):
    _loop_factory = TrioEventLoop


asyncio.set_event_loop_policy(TrioPolicy())
if not isinstance(asyncio.get_event_loop(), TrioEventLoop):  # pragma: no cover
    raise ImportError("You imported trio_asyncio too late.")



from tests import aiotest

class ThreadTests(aiotest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(ThreadTests, cls).setUpClass()
        if not cls.config.support_threads:
            raise aiotest.unittest.SkipTest("threads are not supported")

    def test_ident(self):
        threading = self.config.threading
        try:
            get_ident = threading.get_ident   # Python 3
        except AttributeError:
            get_ident = threading._get_ident   # Python 2

        result = {'ident': None}

        def work():
            result['ident'] = get_ident()

        fut = self.loop.run_in_executor(None, work)
        self.loop.run_until_complete(fut)

        # ensure that work() was executed in a different thread
        work_ident = result['ident']
        self.assertIsNotNone(work_ident)
        self.assertNotEqual(work_ident, get_ident())

    def test_run_twice(self):
        result = []

        def work():
            result.append("run")

        fut = self.loop.run_in_executor(None, work)
        self.loop.run_until_complete(fut)
        self.assertEqual(result, ["run"])

        # ensure that run_in_executor() can be called twice
        fut = self.loop.run_in_executor(None, work)
        self.loop.run_until_complete(fut)
        self.assertEqual(result, ["run", "run"])

    def test_policy(self):
        asyncio = self.config.asyncio
        result = {'loop': 'not set'}   # sentinel, different than None

        def work():
            try:
                result['loop'] = asyncio.get_event_loop()
            except Exception as exc:
                result['loop'] = exc

        # get_event_loop() must return None in a different thread
        fut = self.loop.run_in_executor(None, work)
        self.loop.run_until_complete(fut)
        self.assertIsInstance(result['loop'], (AssertionError, RuntimeError))

    def test_run_in_thread(self):
        asyncio = self.config.asyncio
        threading = self.config.threading

        class LoopThread(threading.Thread):
            def __init__(self, event):
                super(LoopThread, self).__init__()
                self.loop = None
                self.event = event

            def run(self):
                self.loop = asyncio.new_event_loop()
                try:
                    self.loop.set_debug(True)
                    asyncio.set_event_loop(self.loop)

                    self.event.set()
                    self.loop.run_forever()
                finally:
                    self.loop.close()
                    asyncio.set_event_loop(None)

        result = []

        # start an event loop in a thread
        event = threading.Event()
        thread = LoopThread(event)
        thread.start()
        event.wait()
        loop = thread.loop

        def func(loop):
            result.append(threading.current_thread().ident)
            loop.stop()

        # call func() in a different thread using the event loop
        tid = thread.ident
        loop.call_soon_threadsafe(func, loop)

        # stop the event loop
        thread.join()
        self.assertEqual(result, [tid])


if __name__ == '__main__':
    import unittest
    unittest.main()



# XX this does not belong here -- b/c it's here, these things only apply to
# the tests in trio/_core/tests, not in trio/tests. For now there's some
# copy-paste...
#
# this stuff should become a proper pytest plugin

import pytest
import inspect

from trio._core.tests.conftest import trio_test, MockClock

import trio
import trio_asyncio
import asyncio


@pytest.fixture(scope="function", autouse=True)
def loop(request):

    loop = trio_asyncio.TrioEventLoop()
    request.instance.loop = loop
    asyncio.set_event_loop(loop)

    yield loop

    # check that the loop really is idle
    #    if loop._q.qsize():
    #        import pdb;pdb.set_trace()
    #    assert loop._q.qsize() == 0
    assert not loop._timers
    assert not list(
        x for x in loop._delayed_calls if not getattr(x, '_cancelled', False)
    )
    if not loop.is_closed():
        loop.run_task(trio._core.wait_all_tasks_blocked)
        loop.close()


# FIXME: split off into a package (or just make part of trio's public
# interface?), with config file to enable? and I guess a mark option too; I
# guess it's useful with the class- and file-level marking machinery (where
# the raw @trio_test decorator isn't enough).
@pytest.hookimpl(tryfirst=True)
def pytest_pyfunc_call(pyfuncitem):
    if inspect.iscoroutinefunction(pyfuncitem.obj):
        pyfuncitem.obj = trio_test(pyfuncitem.obj)



from __future__ import absolute_import
from tests.aiotest import socketpair
from tests import aiotest

class AddReaderTests(aiotest.TestCase):
    def test_add_reader(self):
        result = {'received': None}
        rsock, wsock = socketpair()
        self.addCleanup(rsock.close)
        self.addCleanup(wsock.close)

        def reader():
            data = rsock.recv(100)
            result['received'] = data
            self.loop.remove_reader(rsock)
            self.loop.stop()

        def writer():
            self.loop.remove_writer(wsock)
            self.loop.call_soon(wsock.send, b'abc')

        self.loop.add_reader(rsock, reader)
        self.loop.add_writer(wsock, writer)

        self.loop.run_forever()
        self.assertEqual(result['received'], b'abc')

    def check_add_replace(self, event):
        socket = self.config.socket

        selector = self.loop._selector
        if event == 'reader':
            add_sock = self.loop.add_reader
            remove_sock = self.loop.remove_reader
            def get_handle(fileobj):
                return selector.get_key(fileobj).data[0]
        else:
            add_sock = self.loop.add_writer
            remove_sock = self.loop.remove_writer
            def get_handle(fileobj):
                return selector.get_key(fileobj).data[1]

        sock = socket.socket()
        self.addCleanup(sock.close)

        def func():
            pass

        def func2():
            pass

        self.assertRaises(KeyError, get_handle, sock)

        add_sock(sock, func)
        handle1 = get_handle(sock)
        self.assertFalse(handle1._cancelled)

        add_sock(sock, func2)
        handle2 = get_handle(sock)
        self.assertIsNot(handle1, handle2)
        self.assertTrue(handle1._cancelled)
        self.assertFalse(handle2._cancelled)

        removed = remove_sock(sock)
        self.assertTrue(removed)
        self.assertTrue(handle2._cancelled)

        removed = remove_sock(sock)
        self.assertFalse(removed)

    def test_add_reader_replace(self):
        self.check_add_replace("reader")

    def test_add_writer_replace(self):
        self.check_add_replace("writer")


if __name__ == '__main__':
    import unittest
    unittest.main()



import unittest
import pytest
import trio_asyncio
import asyncio
import os
import trio
from tests import aiotest


class Seen:
    flag = 0


class MiscTests(aiotest.TestCase):
    def test_close_no_stop(self):
        def close_no_stop():
            self.loop.close()
        self.loop.call_soon(close_no_stop)

        with pytest.raises(RuntimeError) as err:
            res = self.loop.run_forever()
        assert "before closing" in err.value.args[0]

    def test_err(self):
        async def err():
            raise RuntimeError("Foo")

        with pytest.raises(RuntimeError) as err:
            res = self.loop.run_task(err)
        assert err.value.args[0] == "Foo"

    def test_err(self):
        owch = 0

        async def nest():
            nonlocal owch
            owch = 1
            raise Exception("should not run")

        async def call_nested():
            with pytest.raises(RuntimeError) as err:
                self.loop.run_task(next)
            assert "already running" in err.value.args[0]

        self.loop.run_task(call_nested)
        assert not owch

    def test_cancel_sleep(self):
        owch = 0

        def do_not_run():
            nonlocal owch
            owch = 1
            raise Exception("should not run")

        async def cancel_sleep():
            h = self.loop.call_later(0.1, do_not_run)
            await trio.sleep(0.05)
            h.cancel()
            await trio.sleep(0.2)

        self.loop.run_task(cancel_sleep)
        assert owch == 0

    def test_restore_fds(self):
        data = b""
        r,w = os.pipe()

        def has_data():
            nonlocal data
            data += os.read(r,99)
        def send_data(x):
            os.write(w,b'hip')
            self.loop.call_soon(self.loop.remove_writer,w)

        async def x1():
            self.loop.add_reader(r,has_data)
        async def x2():
            self.loop.add_writer(w,send_data)
        async def x3():
            await asyncio.sleep(0.2, loop=self.loop)
            assert not self.loop.remove_writer(r)
            assert self.loop.remove_reader(r)
        async def x4():
            assert not self.loop.remove_reader(r)

        self.loop.run_until_complete(x1())
        self.loop.run_until_complete(x2())
        self.loop.run_until_complete(x3())
        self.loop.run_until_complete(x4())




from tests import aiotest
import signal

class CallbackTests(aiotest.TestCase):
    def test_call_soon(self):
        result = []

        def hello_world(loop):
            result.append('Hello World')
            loop.stop()

        self.loop.call_soon(hello_world, self.loop)
        self.loop.run_forever()
        self.assertEqual(result, ['Hello World'])

    def test_call_soon_control(self):
        result = []

        def func(result, loop):
            loop.call_soon(append_result, loop, result, "yes")
            result.append(str(result))

        def append_result(loop, result, value):
            result.append(value)
            loop.stop()

        self.loop.call_soon(func, result, self.loop)
        self.loop.run_forever()
        # http://bugs.python.org/issue22875: Ensure that call_soon() does not
        # call append_result() immediatly, but when control returns to the
        # event loop, when func() is done.
        self.assertEqual(result, ['[]', 'yes'])

    def test_soon_stop_soon(self):
        result = []

        def hello():
            result.append("Hello")

        def world():
            result.append("World")
            self.loop.stop()

        self.loop.call_soon(hello)
        self.loop.stop()
        self.loop.call_soon(world)

        if False: # self.config.stopping:
            self.loop.run_forever()
            self.assertEqual(result, ["Hello", "World"])
        else:
            self.loop.run_forever()
            # ensure that world() is not called, since stop() was scheduled
            # before call_soon(world)
            self.assertEqual(result, ["Hello"])

            self.loop.run_forever()
            self.assertEqual(result, ["Hello", "World"])

    def test_close(self):
        config = self.config
        if not config.call_soon_check_closed:
            # http://bugs.python.org/issue22922 not implemented
            self.skipTest("call_soon() doesn't raise if the event loop is closed")

        self.loop.close()

        @config.asyncio.coroutine
        def test():
            pass

        func = lambda: False
        coro = test()
        self.addCleanup(coro.close)

        # operation blocked when the loop is closed
        with self.assertRaises(RuntimeError):
            self.loop.run_forever()
        with self.assertRaises(RuntimeError):
            fut = config.asyncio.Future(loop=self.loop)
            self.loop.run_until_complete(fut)
        with self.assertRaises(RuntimeError):
            self.loop.call_soon(func)
        with self.assertRaises(RuntimeError):
            self.loop.call_soon_threadsafe(func)
        with self.assertRaises(RuntimeError):
            self.loop.call_later(1.0, func)
        with self.assertRaises(RuntimeError):
            self.loop.call_at(self.loop.time() + .0, func)
        with self.assertRaises(RuntimeError):
            self.loop.run_in_executor(None, func)
        with self.assertRaises(RuntimeError):
            self.loop.create_task(coro)
        with self.assertRaises(RuntimeError):
            self.loop.add_signal_handler(signal.SIGTERM, func)


if __name__ == '__main__':
    import unittest
    unittest.main()



from tests import aiotest

async def hello_world(asyncio, result, delay):
    result.append("Hello")
    # retrieve the event loop from the policy
    await asyncio.sleep(delay)
    result.append('World')
    return "."

class CoroutineTests(aiotest.TestCase):
    def test_hello_world(self):
        result = []
        coro = hello_world(self.config.asyncio, result, 0.001)
        self.loop.run_until_complete(coro)
        self.assertEqual(result, ['Hello', 'World'])

    def test_waiter(self):
        async def waiter(asyncio, hello_world, result):
            loop = asyncio.get_event_loop()
            fut = asyncio.Future(loop=loop)
            loop.call_soon(fut.set_result, "Future")

            value = await fut
            result.append(value)

            value = await hello_world(asyncio, result, 0.001)
            result.append(value)

        result = []
        coro = waiter(self.config.asyncio, hello_world, result)
        self.loop.run_until_complete(coro)
        self.assertEqual(result, ['Future', 'Hello', 'World', '.'])


if __name__ == '__main__':
    import unittest
    unittest.main()



# This code implements a clone of the asyncio mainloop which hooks into
# Trio.

import sys
import trio
import asyncio
import math
import heapq
import signal
import threading
import selectors

import logging
logger = logging.getLogger(__name__)

from functools import wraps
from asyncio.events import _format_callback, _get_function_source
from selectors import _BaseSelectorImpl, EVENT_READ, EVENT_WRITE

__all__ = ['trio2aio','aio2trio']


def trio2aio(loop='loop'):
    def adapter(proc):
        @wraps(proc)
        async def call(self, *args, **kwargs):
            loop_ = getattr(self, loop)
            return await loop_.call_asyncio(proc, self, *args, **kwargs)
        return call
    if callable(loop):
        proc,loop = loop,'loop'
        return adapter(proc)
    else:
        return adapter


def aio2trio(loop='loop'):
    def adapter(proc):
        @wraps(proc)
        async def call(self, *args, **kwargs):
            loop_ = getattr(self, loop)
            return await loop_.call_trio(proc, self, *args, **kwargs)
        return call
    if callable(loop):
        proc,loop = loop,'loop'
        return adapter(proc)
    else:
        return adapter




import datetime
from tests import aiotest

class TimerTests(aiotest.TestCase):
    def test_display_date(self):
        result = []
        delay = 0.050
        count = 3

        def display_date(end_time, loop):
            if not end_time:
                end_time.append(self.loop.time() + delay * count)
            result.append(datetime.datetime.now())
            if (loop.time() + delay) < end_time[0]:
                loop.call_later(delay, display_date, end_time, loop)
            else:
                loop.stop()

        self.loop.call_soon(display_date, [], self.loop)
        self.loop.run_forever()

        self.assertEqual(len(result), count, result)

    def test_later_stop_later(self):
        result = []

        def hello():
            result.append("Hello")

        def world(loop):
            result.append("World")
            loop.stop()

        self.loop.call_later(0.001, hello)
        self.loop.call_later(0.025, self.loop.stop)
        self.loop.call_later(0.050, world, self.loop)
        self.loop.run_forever()

        self.config.sleep(0.100)
        self.assertEqual(result, ["Hello"])

        self.loop.run_forever()
        self.assertEqual(result, ["Hello", "World"])


if __name__ == '__main__':
    import unittest
    unittest.main()
